(* topologyScript.sml - Formalization of topology (WORK IN PROGRESS)

Copyright © Mario Castelán Castro.

This work is licensed as free software under the GNU General Public License
either version 3 or any later version published by the Free Software
Foundation. See <https://www.gnu.org/licenses/gpl.html>.

As an additional permission (see section 7 of the GNU General Public
License, version 3) you may convey copies of this work or a work based on
this work without a copy of the license. The other requirements of the
license remain in effect. *)

open HolKernel Parse boolLib bossLib;

open marioxcc_tactics;

val _ = new_theory "topology";

(*** Bibliography ***)

(* The following are the main works consulted to find relevant theorems to
formalize. In general, the informal proofs presented therein are not
closely followed, but they may be used as a guide or inspiration for the
formal proof.

[Kelley] Kelley, John L. “General Topology” (1975).
[Mendelson] Mendelson, Bert. “Introduction to topology” (1963).
[Munkres] Munkres, James. “Topology” (2014). *)


(*** Aliases and generic configuration ***)

val _ = ParseExtras.temp_tight_equality ();

val resq_SS = res_quanTools.resq_SS;
val resq_ss = res_quanTools.resq_ss;
val DNF_ss = boolSimps.DNF_ss;
val SET_SPEC_ss = pred_setSimps.SET_SPEC_ss;
val PRED_SET_ss = pred_setSimps.PRED_SET_ss;
val PRED_SET_AC_ss = pred_setSimps.PRED_SET_AC_ss;

val set_ext = pred_setTheory.EXTENSION;
val set_spec = pred_setTheory.SPECIFICATION;
val set_subset = pred_setTheory.SUBSET_DEF;
val set_diff = pred_setTheory.IN_DIFF;
val set_diff_subset = pred_setTheory.DIFF_SUBSET;
val set_union = pred_setTheory.IN_UNION;
val set_inter = pred_setTheory.IN_INTER;
val set_bigunion = pred_setTheory.IN_BIGUNION;
val set_biginter = pred_setTheory.IN_BIGINTER;
val set_compl = pred_setTheory.IN_COMPL;

val disch_tac = DISCH_TAC;


(*** Set theory background ***)

fun unicode_version term_name string =
  (* The “Unicode” structure is from “/src/parse/Parse.sig” in the HOL4
  source code tree. *)
  Unicode.unicode_version { tmnm = term_name, u = string };

unicode_version "DIFF" "−−"; (* U+2212, U+2212 *)
unicode_version "BIGUNION" "⋃"; (* U+22C3 *)
unicode_version "BIGINTER" "⋂"; (* U+22C2 *)

val RES_EXISTS_NOT_EMPTY = store_thm(
  "RES_EXISTS_NOT_EMPTY",
  “∀X P. RES_EXISTS X P ⇒ X ≠ ∅”,
  NTAC 2 GEN_TAC
  >> ONCE_REWRITE_TAC [MONO_NOT_EQ]
  >> ONCE_REWRITE_TAC [satTheory.NOT_NOT]
  >> DISCH_THEN SUBST1_TAC
  >> ACCEPT_TAC (SPEC “P:(α -> bool)” res_quanTheory.RES_EXISTS_EMPTY));

val SUBSET_ELEMENT = store_thm(
  "SUBSET_ELEMENT",
  “∀X Y. X ⊆ Y ⇒ ∀r. r ∈ X ⇒ r ∈ Y”,
  GEN_TAC >> GEN_TAC >> DISCH_TAC
  >> IMP_RES_TAC pred_setTheory.SUBSET_DEF);

val SING_SUBSET = store_thm(
  "SING_SUBSET",
  “∀X r. {r} ⊆ X ⇔ r ∈ X”,
  SIMP_TAC bool_ss [pred_setTheory.SUBSET_DEF, pred_setTheory.IN_SING]);

val SUBSET_REWRITES = store_thm(
  "SUBSET_REWRITES",
  “∀t s. s ⊆ t ⇒
     s ∩ t = s ∧ t ∩ s = s ∧ t ∪ s = t ∧ s ∪ t = t ∧ s −− t = ∅”,
  (* Alternative proof (around 10 times as many primitive inferences):
     PROVE_TAC [
       pred_setTheory.UNION_COMM,
       pred_setTheory.SUBSET_UNION_ABSORPTION,
       pred_setTheory.INTER_COMM,
       pred_setTheory.SUBSET_INTER_ABSORPTION,
       pred_setTheory.SUBSET_DIFF_EMPTY] *)
  GEN_TAC >> GEN_TAC >> DISCH_TAC
  >> ‘s ∩ t = s ∧ s ∪ t = t ∧ s −− t = ∅’ suffices_by (
    ASM_SIMP_TAC (bool_ss ++ PRED_SET_AC_ss) [])
  >> ASM_SIMP_TAC (bool_ss) [
    GSYM pred_setTheory.SUBSET_UNION_ABSORPTION,
    GSYM pred_setTheory.SUBSET_INTER_ABSORPTION,
    pred_setTheory.SUBSET_DIFF_EMPTY]);

val SUBSET_UNION_INTER = store_thm(
  "SUBSET_UNION_INTER",
  “∀X Y Z. (X ⊆ Z ∧ Y ⊆ Z) ⇒ (X ∩ Y ⊆ Z ∧ X ∪ Y ⊆ Z)”,
  SRW_TAC [DNF_ss] [
    pred_setTheory.SUBSET_DEF,
    pred_setTheory.IN_INTER,
    pred_setTheory.IN_UNION]);

(* “pred_set” contains a theorem called “DIFF_UNION” but it is different
from the proposition included in “DIFF_RIGHT_CLAUSES”. *)
val DIFF_RIGHT_CLAUSES = store_thm(
  "DIFF_RIGHT_CLAUSES",
  “∀X Y Z.
     Z −− (X ∪ Y) = (Z −− X) ∩ (Z −− Y) ∧
     Z −− (X ∩ Y) = (Z −− X) ∪ (Z −− Y)”,
  (* “rpt STRIP_TAC” generates a case split here. *)
  rpt STRIP_TAC
  >> REWRITE_TAC [
    pred_setTheory.EXTENSION,
    pred_setTheory.IN_DIFF,
    pred_setTheory.IN_UNION,
    pred_setTheory.IN_INTER]
  >> DECIDE_TAC);

(* THEOREM. Let sp be any set; then the complement relative to sp behaves
as its own inverse for any subset of sp. This does not hold for non-subsets
of sp because taking the complement of any set X relative to sp removes the
elements of X outside sp. *)
val DIFF_SELF_INV = store_thm(
  "DIFF_SELF_INV",
  “∀sp X. X ⊆ sp ⇒ sp −− (sp −− X) = X”,
  rw [pred_setTheory.EXTENSION]
  >> PROVE_TAC [pred_setTheory.SUBSET_DEF]);

val INTER_DIFF_EMPTY = store_thm(
  "INTER_DIFF_EMPTY",
  “∀X Y. X ∩ (Y −− X) = ∅”,
  REWRITE_TAC [
    pred_setTheory.EXTENSION,
    pred_setTheory.IN_INTER,
    pred_setTheory.IN_DIFF,
    pred_setTheory.NOT_IN_EMPTY]
  >> DECIDE_TAC);

val INTER_DIFF = store_thm(
  "INTER_DIFF",
  “∀X Y Z. (X −− Z) ∩ Y = X ∩ Y −− Z”,
  SIMP_TAC bool_ss [
    AC CONJ_ASSOC CONJ_COMM,
    pred_setTheory.EXTENSION,
    pred_setTheory.IN_DIFF,
    pred_setTheory.IN_INTER]);

val DIFF_SING_NOT_ELEMENT = store_thm(
  "DIFF_SING_NOT_ELEMENT",
  “∀X r. r ∉ X ⇒ X −− {r} = X”,
  ASM_SIMP_TAC bool_ss [
  GSYM pred_setTheory.DELETE_DEF,
  pred_setTheory.DELETE_NON_ELEMENT_RWT]);

val SUBSET_DIFF = store_thm(
  "SUBSET_DIFF",
  “∀X Y sp. X ⊆ sp ∧ Y ⊆ sp ⇒ (X ⊆ Y ⇔ sp −− Y ⊆ sp −− X)”,
  strip_all >>
  full_bsimp [set_ext, set_inter, set_subset, set_compl, set_diff] >>
  PROVE_TAC []);

val DIFF_INJ = store_thm(
  "DIFF_INJ",
  “∀X Y sp. X ⊆ sp ∧ Y ⊆ sp ⇒ (sp −− X = sp −− Y ⇔ X = Y)”,
  strip_all >>
  eq_tac
  >- must_prove ‘sp −− X = sp −− Y ⇒ X = Y’ (
    disch_tac >>
    lemma ‘sp −− (sp −− X) = sp −− (sp −− Y)’ (asm_bsimp []) >>
    with_assuml [‘X ⊆ sp’, ‘Y ⊆ sp’, ‘sp −− (sp −− X) = sp −− (sp −− Y)’] >>
    full_bsimp [DIFF_SELF_INV])
  (* This case is trivial. *)
  >- must_prove ‘X = Y ⇒ sp −− X = sp −− Y’ (
    bsimp []));

val DIFF_COMPL = store_thm(
  "DIFF_COMPL",
  “∀X Y. X −− Y = X ∩ COMPL Y”,
  strip_all >>
  bsimp [set_ext, set_diff, set_inter, set_compl]);

val DE_MORGAN_SET = store_thm(
  "DE_MORGAN_SET",
  “∀X Y.
     COMPL (X ∩ Y) = COMPL X ∪ COMPL Y ∧
     COMPL (X ∪ Y) = COMPL X ∩ COMPL Y”,
  NTAC 2 GEN_TAC >>
  bsimp [set_ext, set_compl, set_union, set_inter]);

(* Definition of the intersection of a collection of sets relative to a set
sp. The justification for this definition is given by the theorem
“REL_BIGINTER_COND”. *)
val REL_BIGINTER_DEF = new_definition(
  "REL_BIGINTER_DEF",
  “rel_biginter sp A = sp ∩ (⋂ A)”);
Unicode.uoverload_on ("⋂′", “rel_biginter”); (* U+22C2, U+2032 *)

(* THEOREM. Let A be a collection of subsets of sp. If A is empty then the
intersection of the membes of A relative to sp is sp. If A is non-empty
then the intersection of the members of A relative to sp is equal to their
general intersection. *)
val REL_BIGINTER_COND = store_thm(
  "REL_BIGINTER_COND",
  “∀sp A. (∀X::A. X ⊆ sp) ⇒
     ⋂′ sp A = if A = ∅ then sp else BIGINTER A”,
  SRW_TAC [resq_SS] [pred_setTheory.EXTENSION, REL_BIGINTER_DEF]
  >> PROVE_TAC [pred_setTheory.SUBSET_DEF]);

val IN_REL_BIGINTER = store_thm(
  "IN_REL_BIGINTER",
  “∀sp A r. r ∈ ⋂′ sp A ⇔ r ∈ sp ∧ ∀X::A. r ∈ X”,
  SRW_TAC [resq_SS] [REL_BIGINTER_DEF]);

val DIFF_REL_BIGINTER = store_thm(
  "DIFF_REL_BIGINTER",
  “∀sp A. sp −− ⋂′ sp A = ⋃ {sp −− X | X ∈ A}”,
  RW_TAC bool_ss [pred_setTheory.EXTENSION]
  >> SRW_TAC [DNF_ss, resq_SS] [
    IN_REL_BIGINTER,
    AC boolTheory.CONJ_COMM boolTheory.CONJ_ASSOC]);

(* Counterpart of pred_setTheory.SUBSET_BIGUNION_I for rel_biginter. *)
val SUBSET_REL_BIGINTER_I = store_thm(
  "SUBSET_REL_BIGINTER_I",
  “∀sp A. ∀X::A. ⋂′ sp A ⊆ X”,
  RW_TAC resq_ss [pred_setTheory.SUBSET_DEF, IN_REL_BIGINTER]);

val DIFF_BIGUNION = store_thm(
  "DIFF_BIGUNION",
  “∀sp A. sp −− ⋃ A = ⋂′ sp {sp −− X | X ∈ A}”,
  RW_TAC bool_ss [pred_setTheory.EXTENSION]
  >> SRW_TAC [DNF_ss, resq_SS] [IN_REL_BIGINTER]
  >> PROVE_TAC []);


(*** Fundamental definitions and immediate theorems ***)

(* When “To” is a topology, “space To” is the set of all points in the
corresponding topological space. *)
val SPACE_DEF = new_definition(
  "SPACE_DEF",
  “space To = ⋃ To”);

val _ = set_fixity "open_in" (Infix (NONASSOC, 450));
val OPEN_IN_DEF = new_definition(
  "OPEN_IN_DEF",
  “X open_in To ⇔ X ∈ To”);

val _ = set_fixity "closed_in" (Infix (NONASSOC, 450));
val CLOSED_IN_DEF = new_definition(
  "CLOSED_IN_DEF",
  “X closed_in To ⇔ X ⊆ space To ∧ space To −− X open_in To”);

(* THEOREM. Any open or closed set is a subset of the topological space. *)
val OPEN_CLOSED_SUBSET_SPACE = store_thm(
  "OPEN_CLOSED_SUBSET_SPACE",
  “∀To X. (X open_in To ⇒ X ⊆ space To) ∧ (X closed_in To ⇒ X ⊆ space To)”,
  RW_TAC bool_ss [
    OPEN_IN_DEF,
    CLOSED_IN_DEF,
    SPACE_DEF,
    pred_setTheory.SUBSET_BIGUNION_I]);

(* THEOREM. For any subset X of the topological space, X is closed if and
only if its complement is open and vice-versa. *)
val OPEN_CLOSED_COMPL_EQ = store_thm(
  "OPEN_CLOSED_COMPL_EQ",
  “∀To X. X ⊆ space To ⇒
    (X open_in To ⇔ space To −− X closed_in To) ∧
    (X closed_in To ⇔ space To −− X open_in To)”,
  rw [CLOSED_IN_DEF, DIFF_SELF_INV]);

(* THEOREM. If X is open, its complement is closed and vice-versa *)
val OPEN_CLOSED_COMPL_IMP = store_thm(
  "OPEN_CLOSED_COMPL_IMP",
  “∀To X.
    (X open_in To ⇒ space To −− X closed_in To) ∧
    (X closed_in To ⇒ space To −− X open_in To)”,
  rw [CLOSED_IN_DEF, OPEN_CLOSED_SUBSET_SPACE, DIFF_SELF_INV]);

(* In the usual definition of a topological space (X, t), t is a topology
and X is the set of all points that belong to a set in t. Given a topology
t, X is uniquely determined. In our formalization, we identify a
topological space with its topology.

We use the following conventions to name variables: “To” is a
topology. “r”, “s”, “t” are points. “X”, “Y”, “Z” are sets of points. “A”,
“B”, “C”, “D” are collections of sets of points. *)
val IS_TOPO_DEF = new_definition(
  "IS_TOPO_DEF",
  “is_topo To ⇔
     (∀A. (∀X::A. X open_in To) ⇒ ⋃ A open_in To) ∧
     (∀A B. A open_in To ∧ B open_in To ⇒ A ∩ B open_in To)”);

(* THEOREM. If X and Y are closed sets, then X ∪ Y is closed. *)
val UNION_CLOSED = store_thm(
  "UNION_CLOSED",
  “∀To. is_topo To ⇒
    ∀X Y. X closed_in To ∧ Y closed_in To ⇒ X ∪ Y closed_in To”,
  SRW_TAC [resq_SS] [CLOSED_IN_DEF, IS_TOPO_DEF, DIFF_RIGHT_CLAUSES]);

(* THEOREM. The intersection (relative to the topological space) of a
collection of closed sets is closed. *)
val REL_BIGINTER_CLOSED = store_thm(
  "REL_BIGINTER_CLOSED",
  “∀To. is_topo To ⇒ ∀A. (∀X::A. X closed_in To) ⇒
    ⋂′ (space To) A closed_in To”,
  rw [CLOSED_IN_DEF, DIFF_REL_BIGINTER] >| [
    rw [REL_BIGINTER_DEF],

    UNDISCH_TAC “is_topo To”
    >> SRW_TAC [DNF_ss, resq_SS] [IS_TOPO_DEF]]);

(* THEOREM. Characterization of topological spaces in terms of closed sets:
To is a topology if and only if the intersection (realtive to the
topological space) of any collection of closed sets is closed and for any
subsets of the topological space X, Y it is the case that X ∩ Y is closed.

We use “sp” as an abbreviation for “space To” in the formal proof and the
accompanying commentary. *)
val TOPO_CLOSED_SETS = store_thm(
  "TOPO_CLOSED_SETS",
  “∀To. is_topo To ⇔
     (∀A. (∀X::A. X closed_in To) ⇒ ⋂′ (space To) A closed_in To) ∧
     (∀X Y. X closed_in To ∧ Y closed_in To ⇒ X ∪ Y closed_in To)”,
  (* Split the biconditional. *)
  GEN_TAC >> EQ_TAC
  (* First implication: From is_topo To follow the asserted properties of
  closed sets. We simply use the already proved theorems. *)
  >- RW_TAC bool_ss [REL_BIGINTER_CLOSED, UNION_CLOSED]
  (* Second implication: From the asserted properties of closed sets
  follows is_topo To. *)
  >> qabbrev_tac ‘sp = space To’
  >> STRIP_TAC
  (* (1) If A is any colletion of open sets, then the union of the elements
  of A is an open set. *)
  >> lemma ‘∀A. (∀X::A. X open_in To) ⇒ ⋃ A open_in To’ (
    GEN_TAC >> DISCH_TAC
    (* (1.1) {sp −− X | X ∈ A} is a collection of closed sets by
    OPEN_CLOSED_COMPL_IMP. *)
    >> lemma ‘∀X::{sp −− X | X ∈ A}. X closed_in To’ (
      POP_ASSUM MP_TAC
      >> SIMP_TAC (bool_ss ++ resq_SS ++ SET_SPEC_ss ++ DNF_ss) []
      >> PROVE_TAC [OPEN_CLOSED_COMPL_IMP])
    (* Therefore ⋂′ sp {sp −− X | X ∈ A} is closed by hypothesis. *)
    >> now ‘⋂′ sp {sp −− X | X ∈ A} closed_in To’ (RW_TAC bool_ss [])
    (* Therefore sp −− ⋃ A is closed by DIFF_BIGUNION. *)
    >> now ‘sp −− ⋃ A closed_in To’ (RW_TAC bool_ss [DIFF_BIGUNION])
    (* (1.2) It was proved already that A is a collection of subsets of
    sp. Thus ⋃ A ⊆ sp follows by OPEN_CLOSED_SUBSET_SPACE. *)
    >> lemma ‘⋃ A ⊆ sp’ (
      qpat_assum ‘∀X::A. X open_in To’ MP_TAC
      >> SIMP_TAC resq_ss []
      >> PROVE_TAC [
        OPEN_CLOSED_SUBSET_SPACE,
        pred_setTheory.BIGUNION_SUBSET])
    (* From (1.1) and (1.2) by OPEN_CLOSED_COMPL_EQ follows that ⋃ A is
    open. *)
    >> RW_TAC bool_ss [OPEN_CLOSED_COMPL_EQ])
  (* (2) If X and Y are open then X ∩ Y is open. *)
  >> lemma ‘∀X Y. X open_in To ∧ Y open_in To ⇒ X ∩ Y open_in To’ (
    NTAC 2 GEN_TAC >> DISCH_TAC
    (* (2.1) sp −− X and sp −− Y are closed by OPEN_CLOSED_COMPL_IMP *)
    >> lemma ‘sp −− X closed_in To ∧ sp −− Y closed_in To’ (
      PROVE_TAC [OPEN_CLOSED_COMPL_IMP])
    (* Therefore (sp −− X) ∪ (sp −− Y) is closed by hypothesis. *)
    >> now ‘(sp −− X) ∪ (sp −− Y) closed_in To’ (RW_TAC bool_ss [])
    (* Therefore sp −− (X ∩ Y) is closed by DIFF_RIGHT_CLAUSES. *)
    >> now ‘sp −− (X ∩ Y) closed_in To’ (
      RW_TAC bool_ss [DIFF_RIGHT_CLAUSES])
    (* (2.2) X and Y are closed by hypothesis. Thus X ∩ Y ⊆ sp follows by
    OPEN_CLOSED_SUBSET_SPACE.  *)
    >> lemma ‘X ∩ Y ⊆ sp’ (
      PROVE_TAC [OPEN_CLOSED_SUBSET_SPACE, SUBSET_UNION_INTER])
    (* From (2.1) and (2.2) by OPEN_CLOSED_COMPL_EQ follows that X ∩ Y
    is open. *)
    >> RW_TAC bool_ss [OPEN_CLOSED_COMPL_EQ])
  (* Use (1) and (2) to prove the second implication. *)
  >> RW_TAC bool_ss [IS_TOPO_DEF]);

val empty_open = prove(
  “∀To. is_topo To ⇒ ∅ open_in To”,
    rw [IS_TOPO_DEF, OPEN_IN_DEF]
    >> ‘(∀X. X ∈ ∅ ⇒ X ∈ To) ⇒ ⋃ ∅ ∈ To’ by (SRW_TAC [resq_SS] [])
    >> fs []);

val space_closed = prove(
  “∀To. is_topo To ⇒ space To closed_in To”,
  rw [CLOSED_IN_DEF, empty_open]);

val space_open = prove(
  “∀To. is_topo To ⇒ space To open_in To”,
  rw [IS_TOPO_DEF, SPACE_DEF, OPEN_IN_DEF]);

val empty_closed = prove(
  “∀To. is_topo To ⇒ ∅ closed_in To”,
  rw [CLOSED_IN_DEF, space_open]);

val TRIVIAL_CLOPEN = store_thm(
  "TRIVIAL_CLOPEN",
  “∀To. is_topo To ⇒
    ∅ open_in To ∧
    ∅ closed_in To ∧
    (space To) open_in To ∧
    (space To) closed_in To”,
  rw [empty_open, empty_closed, space_open, space_closed]);


(*** Basis and subbasis ***)

val IS_BASIS_DEF = new_definition(
  "IS_BASIS_DEF",
  “is_basis A ⇔ ∀X Y::A. X ∩ Y ∈ A”);

val TOPO_FROM_BASIS_DEF = new_definition(
  "TOPO_FROM_BASIS_DEF",
  “topo_from_basis B = {X | ∀r::X. ∃Y::B. r ∈ Y}”);

(* THEOREM. The function topo_from_basis when applied to a basis yields a
topology. *)
val TOPO_FROM_BASIS = store_thm(
  "TOPO_FROM_BASIS",
  “∀A. is_basis A ⇒ is_topo (topo_from_basis A)”,
  SRW_TAC [resq_SS] [IS_BASIS_DEF, IS_TOPO_DEF, TOPO_FROM_BASIS_DEF, OPEN_IN_DEF]
  >> PROVE_TAC []);


(*** Interior points, limit points and closure. ***)

(* The following is a summary of the approaches used in our references to
define the interior and closure of a set X. These are paraphrasis, not
literal quotes.

[Mendelson] Closure: The sets of points for which each neighborhood N we
have N ∩ X ≠ ∅. Interior: The set of points that are element of an open
subset of X.

[Munkres] Closure: The intersection of all closed supersets of X. Interior:
The union of all open subsets of X.

[Kelley] Closure: Same as in [Munkres]. Interior: Same as in [Mendelson].

The definition of [Munkres] is elegant for its symmetry.

In this formalization, the interior of a set is given by partially applying
“interior” (because of currying). *)


(* DEFINITION. Let X be a set. r is an interior point X (in the topology
To) if an only if there exists an open set Y such that r ∈ Y ⊆ X.

In our informal commentary we use the following convention: When To is a
topology we say that X is a neighborhood of r if and only if x is an
interior point of X. Note that given a topology the relation “is a
neighborhood” is the inverse of the relation “is an interior point”. *)
val INTERIOR_DEF = new_definition(
  "INTERIOR_DEF",
  “interior To X r ⇔
     is_topo To ∧ X ⊆ space To ∧ r ∈ space To ∧
     ∃Y. r ∈ Y ∧ Y ⊆ X ∧ Y open_in To”);

(* DEFINITION. Let X be a set. r is a limit point of X if and only if each
neighborhood of r intersects X in some point other than r.

The name “set_limit” is used to avoid confusion with limits of sequences
and nets. *)
val SET_LIMIT_DEF = new_definition(
  "SET_LIMIT_DEF",
  “set_limit To X r ⇔
     is_topo To ∧ X ⊆ space To ∧ r ∈ space To ∧
     ∀Y. interior To Y r ⇒ (X ∩ Y) −− {r} ≠ ∅”);

val PNT_INTERIOR_ELEMENT = store_thm(
  "PNT_INTERIOR_ELEMENT",
  “∀To. is_topo To ⇒ ∀X r. interior To X r ⇒ r ∈ X”,
  GEN_TAC >> DISCH_TAC >> NTAC 2 GEN_TAC
  >> REWRITE_TAC [INTERIOR_DEF]
  >> DISCH_THEN STRIP_ASSUME_TAC
  >> IMP_RES_TAC pred_setTheory.SUBSET_DEF);

(* THEOREM. Let X and Y be subsets of a topological space. If X ⊆ Y, then
every interior point of X is a limit point of Y. *)
val PNT_INTERIOR_SUBSET = store_thm(
  "PNT_INTERIOR_SUBSET",
  “∀To X Y. is_topo To ∧ X ⊆ space To ∧ Y ⊆ space To ∧ X ⊆ Y ⇒
    ∀r. interior To X r ⇒ interior To Y r”,
  rpt STRIP_TAC
  >> must_prove ‘interior To Y r’ (
    ‘r ∈ space To ∧ ∃Z. r ∈ Z ∧ Z ⊆ X ∧ Z open_in To’ by (
      use ‘interior To X r’
      >> SIMP_TAC bool_ss [INTERIOR_DEF])
    >> lemma ‘Z ⊆ Y’ (
      ‘Z ⊆ X ∧ X ⊆ Y’ suffices_by
        MATCH_ACCEPT_TAC pred_setTheory.SUBSET_TRANS
      >> ASM_SIMP_TAC bool_ss [])
    >> ASM_SIMP_TAC bool_ss [INTERIOR_DEF]
    >> qexists_tac ‘Z’
    >> ASM_SIMP_TAC bool_ss []));

(* THEOREM. The interior of a set is equal to the union of all its open
subsets. *)
val INTERIOR_UNION_OPEN = store_thm(
  "INTERIOR_UNION_OPEN",
  “∀To. is_topo To ⇒ ∀X. X ⊆ space To ⇒
     interior To X = ⋃ {Y | Y ⊆ X ∧ Y open_in To}”,
  strip_all >>
  suffices ‘∀r. interior To X r ⇔ r ∈ ⋃ {Y | Y ⊆ X ∧ Y open_in To}’ (
    bsimp [set_spec, set_ext]) >>
  gen_tac >>
  ASM_SIMP_TAC (bool_ss ++ SET_SPEC_ss) [
    INTERIOR_DEF,
    set_ext,
    set_bigunion] >>
  eq_tac >| [
    (* Trivial subgoal *)
    bsimp [],
    (* Trivial set algebra. *)
    bsimp [] >>
    strip_tac >>
    must_prove ‘r ∈ space To’ (
      with_assuml [‘s ⊆ X’, ‘X ⊆ space To’, ‘r ∈ s’] >>
      full_bsimp [set_subset])]);

(* THEOREM. The interior of a set in a topological space is open. *)
val INTERIOR_OPEN = store_thm(
  "INTERIOR_OPEN",
  “∀To. is_topo To ⇒ ∀X. X ⊆ space To ⇒ interior To X open_in To”,
  SIMP_TAC (bool_ss ++ PRED_SET_ss) [IS_TOPO_DEF, INTERIOR_UNION_OPEN]);

(* THEOREM. The interior of a set X is the largest open subset of X. *)
val INTERIOR_IS_LARGEST_OPEN_SUBSET = store_thm(
  "INTERIOR_IS_LARGEST_OPEN_SUBSET",
  “∀To. is_topo To ⇒ ∀X Y. X ⊆ space To ∧ Y ⊆ space To ⇒ Y ⊆ X ⇒
     Y open_in To ⇒ Y ⊆ interior To X”,
  (* Strangely, SUBSET_BIGUNION_I is not closed (in the sense that it has
  free variables). Note that this does not prevent the rewrite from
  working. *)
  SIMP_TAC (bool_ss ++ PRED_SET_ss)
    [IS_TOPO_DEF, INTERIOR_UNION_OPEN, pred_setTheory.SUBSET_BIGUNION_I]);

(* THEOREM. For any subset X of a topological space, X is open if and only
if the elements of X are exactly the interior points of X. *)
val OPEN_PNT_INTERIOR = store_thm(
  "OPEN_PNT_INTERIOR",
  “∀To. is_topo To ⇒
    ∀X. (X open_in To ⇔ (∀r. interior To X r ⇔ r ∈ X))”,
  GEN_TAC >> DISCH_TAC >> GEN_TAC
  (* Split the biconditional *)
  >> EQ_TAC >| [
    (* Case 1 *)
    must_prove ‘X open_in To ⇒ (∀r. interior To X r ⇔ r ∈ X)’ (
      (* PROVE_TAC [INTERIOR_DEF, OPEN_CLOSED_SUBSET_SPACE,
      pred_setTheory.SUBSET_DEF] solves the goal outright but for
      efficiency we do some preprocessing before the first order
      reasoner. *)
      DISCH_TAC >> GEN_TAC
      >> lemma ‘X ⊆ space To’ (IMP_RES_TAC OPEN_CLOSED_SUBSET_SPACE)
      >> REWRITE_TAC [INTERIOR_DEF]
      >> PROVE_TAC [pred_setTheory.SUBSET_DEF]),
   (* Case 2 *)
   must_prove ‘(∀r. interior To X r ⇔ r ∈ X) ⇒ X open_in To’ (
     (* First prove that X is equal to the union of all its open subsets,
     then we prove that this union is open. *)
     DISCH_TAC
     >> qabbrev_tac ‘A = {Y | Y ⊆ X ∧ Y open_in To}’
     >> qabbrev_tac ‘Z = ⋃ A’
     >> lemma ‘X = Z’ (
       SIMP_TAC (bool_ss ++ SET_SPEC_ss) [
         Abbr ‘Z’,
         Abbr ‘A’,
         pred_setTheory.EXTENSION,
         pred_setTheory.IN_BIGUNION]
       >> PROVE_TAC [INTERIOR_DEF, pred_setTheory.SUBSET_DEF])
     >> lemma ‘∀Y::A. Y open_in To’ (
       SIMP_TAC (resq_ss ++ SET_SPEC_ss) [Abbr ‘A’])
     >> must_prove ‘X open_in To’ (
       qpat_assum ‘is_topo To’ MP_TAC
       >> RW_TAC bool_ss [Abbr ‘Z’, IS_TOPO_DEF]))]);

(* Variation of OPEN_PNT_INTERIOR. *)
val OPEN_PNT_INTERIOR_SET = store_thm(
  "OPEN_PNT_INTERIOR_SET",
  “∀To. is_topo To ⇒ ∀X. X open_in To ⇔ interior To X = X”,
  MP_TAC OPEN_PNT_INTERIOR
  >> REWRITE_TAC [boolTheory.FUN_EQ_THM, pred_setTheory.SPECIFICATION]);

(* THEOREM. A subset X of a topological space is open if and only if X is a
neighborhood of each of its points. *)
val OPEN_PNT_INTERIOR_ALT = store_thm(
  "OPEN_PNT_INTERIOR_ALT",
  “∀To. is_topo To ⇒
    ∀X. X open_in To ⇔ (∀r::X. interior To X r)”,
  res_quanTools.RESQ_TAC
  >> PROVE_TAC [OPEN_PNT_INTERIOR, PNT_INTERIOR_ELEMENT]);

val NOT_PNT_LIMIT = store_thm(
  "NOT_PNT_LIMIT",
  “∀To. is_topo To ⇒ ∀X. X ⊆ space To ⇒ ∀r::space To.
     ¬set_limit To X r ⇔ ∃Y. interior To Y r ∧ X ∩ Y −− {r} = ∅”,
  SIMP_TAC (resq_ss) [SET_LIMIT_DEF]);

(* THEOREM. A subset X of a topological space is a neighborhood of r if and
only if r ∈ X and r is not a limit point of the complement of X. *)
val PNT_INTERIOR_PNT_LIMIT = store_thm(
  "PNT_INTERIOR_PNT_LIMIT",
  “∀To. is_topo To ⇒ ∀X. X ⊆ space To ⇒ ∀r.
    interior To X r ⇔ r ∈ X ∧ ¬set_limit To (space To −− X) r”,
  rpt STRIP_TAC
  >> lemma ‘(space To −− X) ⊆ space To’ (
    MATCH_ACCEPT_TAC pred_setTheory.DIFF_SUBSET)
  >> lemma ‘r ∈ X ⇒ r ∈ space To’ (
    with_assum ‘X ⊆ space To’
    >> full_bsimp [pred_setTheory.SUBSET_DEF])
  >> EQ_TAC
  (* Case 1 *)
  >- must_prove
    ‘interior To X r ⇒ r ∈ X ∧ ¬set_limit To (space To −− X) r’ (
    DISCH_TAC
    >> lemma ‘r ∈ X’ (IMP_RES_TAC PNT_INTERIOR_ELEMENT)
    >> lemma ‘r ∈ space To’ (
      with_assuml [‘r ∈ X’, ‘r ∈ X ⇒ r ∈ space To’]
      >> full_bsimp [pred_setTheory.SUBSET_DEF])
    >> ASM_SIMP_TAC (bool_ss ++ resq_SS) [NOT_PNT_LIMIT]
    >> qexists_tac ‘X’
    >> ASM_SIMP_TAC (bool_ss ++ PRED_SET_AC_ss) [
      INTER_DIFF_EMPTY,
      pred_setTheory.EMPTY_DIFF])
  (* Case 2 *)
  >- must_prove
    ‘r ∈ X ∧ ¬set_limit To (space To −− X) r ⇒ interior To X r’ (
    STRIP_TAC
    >> ‘∃Y. interior To Y r ∧ (space To −− X) ∩ Y −− {r} = ∅’ by (
      use ‘¬set_limit To (space To −− X) r’
      >> ASM_SIMP_TAC (bool_ss ++ resq_SS) [NOT_PNT_LIMIT])
    (* Lemma needed for the proof of X ⊆ Y and to apply
    PNT_INTERIOR_SUBSET. *)
    >> lemma ‘Y ⊆ space To’ (
     use ‘interior To Y r’
     >> bsimp [INTERIOR_DEF])
    (* Rewrite expression of neighborhood of r with set algebra. *)
    >> lemma ‘Y ⊆ X’ (
      ‘Y −− X = ∅’ suffices_by (
        bsimp [pred_setTheory.SUBSET_DIFF_EMPTY])
      >> lemma ‘(Y ∩ space To) −− ({r} ∪ X) = ∅’ (
        use ‘(space To −− X) ∩ Y −− {r} = ∅’
        >> SIMP_TAC (bool_ss ++ PRED_SET_AC_ss) [
          INTER_DIFF,
          GSYM pred_setTheory.DIFF_UNION])
      >> lemma ‘{r} ⊆ X’ (
        asm_bsimp [SING_SUBSET])
      >> use ‘Y ∩ space To −− ({r} ∪ X) = ∅’
      >> asm_bsimp [SUBSET_REWRITES])
    >> must_prove ‘interior To X r’ (
      use ‘interior To Y r’
      >> MATCH_MP_TAC PNT_INTERIOR_SUBSET
      >> asm_bsimp [])));

(* Restatement of “PNT_INTERIOR_PNT_LIMIT” in term of sets. *)
val PNT_INTERIOR_PNT_LIMIT_SET = store_thm(
  "PNT_INTERIOR_PNT_LIMIT_SET",
  “∀To. is_topo To ⇒ ∀X. X ⊆ space To ⇒
     interior To X = X −− set_limit To (space To −− X)”,
  (* This is a simple proof based on rewriting until
  “PNT_INTERIOR_PNT_LIMIT” can be used to solve the goal. *)
  strip_all >>
  suffices ‘∀r. r ∈ interior To X ⇔
              r ∈ X −− set_limit To (space To −− X)’ (
    bsimp [set_ext]) >>
  gen_tac >>
  suffices ‘interior To X r ⇔ r ∈ X ∧ ¬set_limit To (space To −− X) r’ (
    bsimp [set_diff, set_spec]) >>
  asm_bsimp [PNT_INTERIOR_PNT_LIMIT]);

(* THEOREM. Any subset X of a topological space is closed if and only if it
contains all its limit points. *)
val CLOSED_PNT_LIMIT_SET = store_thm(
  "CLOSED_PNT_LIMIT_SET",
  “∀To. is_topo To ⇒ ∀X. X ⊆ space To ⇒
    (X closed_in To ⇔ set_limit To X ⊆ X)”,
  strip_all >>
  qabbrev_tac ‘sp = space To’ >>
  (* Eliminate “closed_in” for “open_in”. *)
  suffices ‘sp −− X open_in To ⇔ set_limit To X ⊆ X’ (
    asm_bsimp [Abbr ‘sp’, OPEN_CLOSED_COMPL_EQ]) >>
  (* Eliminate “open_in” to “interior”. *)
  suffices ‘interior To (sp −− X) = sp −− X ⇔ set_limit To X ⊆ X’ (
    asm_bsimp [OPEN_PNT_INTERIOR_SET]) >>
  (* Eliminate “interior” to “set_limit”. This is the key step. The heavy
  work has already been done in PNT_INTERIOR_PNT_LIMIT. *)
  suffices ‘sp −− (X ∪ set_limit To X) = sp −− X ⇔ set_limit To X ⊆ X’ (
    lemma ‘sp −− X ⊆ sp’ (bsimp [set_diff_subset]) >>
    asm_bsimp [
      Abbr ‘sp’,
      DIFF_SELF_INV,
      PNT_INTERIOR_PNT_LIMIT_SET,
      pred_setTheory.DIFF_UNION]) >>
  (* The rest is just set algebra to prove both sides equivalent. *)
  suffices ‘set_limit To X ∪ X = X ⇔ set_limit To X ⊆ X’ (
    lemma ‘set_limit To X ⊆ sp’ (
      suffices ‘∀r. set_limit To X r ⇒ r ∈ sp’ (
        bsimp [set_spec, set_subset]) >>
      asm_bsimp [Abbr ‘sp’, SET_LIMIT_DEF]) >>
    lemma ‘X ∪ set_limit To X ⊆ sp’ (
      with_assuml [‘X ⊆ sp’, ‘set_limit To X ⊆ sp’] >>
      asm_bsimp [pred_setTheory.UNION_SUBSET]) >>
    ASM_SIMP_TAC (bool_ss ++ PRED_SET_AC_ss) [DIFF_INJ]) >>
  bsimp [pred_setTheory.SUBSET_UNION_ABSORPTION]);

val CLOSED_PNT_LIMIT = store_thm(
  "CLOSED_PNT_LIMIT",
  “∀To. is_topo To ⇒ ∀X. X ⊆ space To ⇒
    (X closed_in To ⇔ ∀r. set_limit To X r ⇒ r ∈ X)”,
  strip_all >>
  suffices ‘(X closed_in To ⇔ set_limit To X ⊆ X)’ (
    bsimp [set_spec, set_ext, set_subset]) >>
  asm_bsimp [CLOSED_PNT_LIMIT_SET]);

(* DEFINITION. The closure of a set X is the intersection of all closed
supersets of X. *)
val CLOSURE_DEF = new_definition(
  "CLOSURE_DEF",
  “closure To X = ⋂′ (space To) {Y | X ⊆ Y ∧ Y closed_in To}”);

(* The closure of a set in a topological space is closed. *)
val CLOSURE_CLOSED = store_thm(
  "CLOSURE_CLOSED",
  “∀To. is_topo To ⇒ ∀X. closure To X closed_in To”,
  RW_TAC (bool_ss ++ SET_SPEC_ss) [CLOSURE_DEF, REL_BIGINTER_CLOSED]);

(* The closure of a set X is the smallest closed superset of X. *)
val CLOSURE_IS_SMALLEST_CLOSED_SUPERSET = store_thm(
  "CLOSURE_IS_SMALLEST_CLOSED_SUPERSET",
  “∀To. is_topo To ⇒ ∀X Y. X ⊆ Y ∧ Y closed_in To ⇒ closure To X ⊆ Y”,
  RW_TAC (bool_ss ++ SET_SPEC_ss) [CLOSURE_DEF, SUBSET_REL_BIGINTER_I]);

val _ = export_theory ();
